# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    "serialize",
    "deserialize",
]

from ._core import *
from .internals import BaseUID, EternalUID, ZefRefUID
from ._ops import *
from .op_structs import ZefOp, CollectingOp, SubscribingOp, ForEachingOp, LazyValue, Awaitable, is_python_scalar_type
from .abstract_raes import Entity, Relation, AttributeEntity
from .error import _ErrorType, Error
from .image import Image
from .fx.fx_types import FXElement, Effect
from .flat_graph import FlatGraph_, FlatRef, FlatRefs
from ..pyzef import internals as pyinternals
from .VT import ValueType_
from .symbolic_expression import SymbolicExpression_

##############################
# * Definition
#----------------------------

def get_serializable_types():
    return list(serialization_mapping.keys())

# This will get filled in soon
serialization_mapping = {}
deserialization_mapping = {}

def is_serialization(v):
    return isinstance(v, dict) and v.get("_zeftype", None) == "serialization"

def serialize(v):
    """
    Given a dictionary, list, and any Zeftype this function converts all of the
    ZefTypes into dictionaries that describe the ZefType and allows it to be
    serialized using JSON dumps.

    This function wraps the result of serialize_internal with a serialization
    header. The output of this function will always be a dictionary, even if
    serialize_internal does no work.
    """

    return {"_zeftype": "serialization",
            "version": 1,
            "data": serialize_internal(v)}

def serialize_internal(v):
    """
    This function is the recursive core of the serialization, which does not
    wrap the objects with any header.
    """

    # Serializes a list and recursively calls itself if one of the list elements is of type List
    if type(v) == bytes:
        # Our target is for a JSON-compatible format, so need to convert this to
        # a string smoehow. We choose base64 encoding.
        import base64
        return {
            "_zeftype": "bytes",
            "data": base64.b64encode(v).decode("utf-8")
        }
    if is_python_scalar_type(v):
        return v
    elif type(v) in serialization_mapping:
         return serialization_mapping[type(v)](v)
    raise Exception(f"Don't know how to serialize type {type(v)}")

def deserialize(v):
    """
    Given an output from a previous call to serialize, convert the contained
    Zeftypes, Dicts, Lists into their original forms. This function recursively
    calls itself and other internal functions to fully deserialize nested lists
    and dictionaries.
    """
    if not is_serialization(v):
        from .logger import log
        log.warn("Warning, deserializing an object without a serialization header. This behaviour is deprecated.")
        return deserialize_internal(v)

    if v.get("version", None) != 1:
        raise Exception("Don't understand serialization version '{v.get('version', None)}'")

    return deserialize_internal(v["data"])

def deserialize_internal(v):
    """
    This function is the recursive core of the deserialization, where the
    objects are not wrapped in the serialization header.
    """
    from .logger import log
    if isinstance(v, dict) and "_zeftype" in v:
        v = deserialization_mapping[v["_zeftype"]](v)
    elif isinstance(v, dict):
        v = deserialize_dict(v)
    elif isinstance(v, list):
        v = deserialize_list(v)
    elif isinstance(v, tuple):
        log.warning("Should never get to the point of trying to deserialize a tuple anymore.")
        v = deserialize_tuple(v)

    return v


####################################
# * Implementations
#----------------------------------
def serialize_flatgraph_or_flatref(fg_or_fr) -> dict:
    if isinstance(fg_or_fr, FlatGraph_):
        return {
            "_zeftype": "FlatGraph",
            "key_dict": serialize_dict(fg_or_fr.key_dict),
            "blobs": serialize_tuple(fg_or_fr.blobs),
        }
    elif isinstance(fg_or_fr, FlatRef):
        return {
            "_zeftype": "FlatRef",
            "fg": serialize_flatgraph_or_flatref(fg_or_fr.fg),
            "idx": fg_or_fr.idx,
        }
    elif isinstance(fg_or_fr, FlatRefs):
        return {
            "_zeftype": "FlatRefs",
            "fg": serialize_flatgraph_or_flatref(fg_or_fr.fg),
            "idxs": serialize_list(fg_or_fr.idxs),
        }

def serialize_tuple(l: tuple) -> dict:
    return {
        "_zeftype": "tuple",
        "items": [serialize_internal(el) for el in l]
    }

def serialize_list(l: list) -> list:
    return [serialize_internal(el) for el in l]

def serialize_dict(json_d: dict) -> dict:
    # return {k: serialize(v) for k,v in json_d.items()}
    return {
        "_zeftype": "dict",
        "items": [[serialize_internal(k), serialize_internal(v)] for k,v in json_d.items()]
    }

def serialize_zeftypes(z) -> dict:
    if isinstance(z, ZefRef):
        return {
                "_zeftype"  : "ZefRef",
                "guid"      : str(base_uid(Graph(z))),
                "uid"       : str(base_uid(z)),
                "tx_uid"    : str(base_uid(z | frame | to_tx)),
            }

    elif isinstance(z, EZefRef):
        return {
                "_zeftype"  : "EZefRef",
                "guid"      : str(base_uid(Graph(z))),
                "uid"       : str(base_uid(z)),
            }

    # elif isinstance(z, ZefRefs) or isinstance(z, EZefRefs):
    #     bt_type = {ZefRefs: "ZefRefs", EZefRefs: "EZefRefs"}[type(z)]
    #     tx_uid = str(base_uid(z | frame | to_tx)) if bt_type == "ZefRefs" else None
    #     guid = str(base_uid(Graph(z | first))) if len(z) > 0 else None
    #     return {
    #         "_zeftype"  : bt_type,
    #         "tx_uid"    : tx_uid,
    #         "guid"      : guid, # could be None if ZefRefs or UZefrefs was empty
    #         "value" : [{"uid": str(base_uid(zr))} for zr in z]
    #             }

    elif isinstance(z, RelationType) or isinstance(z, EntityType) or isinstance(z, AttributeEntityType):
        bt_type = {RelationType: "RT", EntityType: "ET", AttributeEntityType: "AET"}[type(z)]
        absorbed_args = LazyValue(z) | absorbed | collect
        absorbed_args = serialize_internal(absorbed_args)
        return {"_zeftype": bt_type, "value": str(z), "absorbed": absorbed_args}

    elif isinstance(z, Graph):
        return {"_zeftype": "Graph", "guid": str(uid(z))}

    elif isinstance(z, ZefEnumValue):
        return {"_zeftype": "Enum", "enum_type": z.enum_type, "enum_val": z.enum_value}

    elif isinstance(z, QuantityFloat) or isinstance(z, QuantityInt):
        q_type = {QuantityFloat: "QuantityFloat", QuantityInt: "QuantityInt"}[type(z)]
        return {"_zeftype": q_type, "value": z.value, "unit": serialize_internal(z.unit)}

    elif isinstance(z, Time):
        return {"_zeftype": "Time", "value": z.seconds_since_1970} 

    elif type(z) in [ZefOp, CollectingOp, SubscribingOp, ForEachingOp]:
        z_type = type(z).__name__
        return serialize_zefops(z_type, z.el_ops)

    elif type(z) in [LazyValue, Awaitable]:
        if isinstance(z, LazyValue):
            additional_dict = {"initial_val": z.initial_val}
        else:
            additional_dict = {"pushable": z.pushable}
        z_type = {LazyValue: "LazyValue", Awaitable: "Awaitable"}[type(z)]
        inner_ztype = {ZefOp: "ZefOp", CollectingOp: "CollectingOp", SubscribingOp: "SubscribingOp", ForEachingOp: "ForEachingOp"}[type(z.el_ops)]

        return {"_zeftype": z_type, "el_ops": serialize_zefops(inner_ztype, z.el_ops.el_ops), **additional_dict}

    elif type(z) in [BaseUID, EternalUID, ZefRefUID]:
        return {"_zeftype": "UID", "value": str(z)}

    elif isinstance(z, Image):
        import zstd, base64
        encoded_buffer = z.buffer
        encoded_buffer = base64.b64encode(zstd.decompress(encoded_buffer)).decode('utf8')
        return {"_zeftype": "Image", "format": z.format, "compression": z.compression, "buffer" : encoded_buffer}

    elif type(z) in [Entity, Relation, AttributeEntity]:
        abstract_type = {Entity: "Entity", Relation: "Relation", AttributeEntity: "AttributeEntity"}[type(z)]
        uid_or_uids = "uids" if abstract_type == "Relation" else "uid"
        type_or_types = [serialize_internal(rae) for rae in z.d['type']] if abstract_type == "Relation" else serialize_internal(z.d['type'])
        absorbed_args = z.d['absorbed']
        return {"_zeftype": abstract_type, "type": type_or_types, uid_or_uids: serialize_internal(z.d[uid_or_uids]), 'absorbed': serialize_internal(absorbed_args)}

    elif isinstance(z, _ErrorType):
        return {"_zeftype": "ErrorType", "type": z.name, "args": serialize_list(z.args)}

    elif isinstance(z, FXElement):
        return {"_zeftype": "FXElement", "elements": [e for e in z.d]}

    else:
        raise NotImplementedError(f"{z} isn't part of the supported deserializable zeftypes!")

def serialize_delegate(z) -> dict:
    if isinstance(z, Delegate):
        return {"_zeftype": "Delegate", "order": z.order, "item": serialize_internal(z.item)}
    elif isinstance(z, pyinternals.DelegateTX):
        return {"_zeftype": "DelegateTX"}
    elif isinstance(z, pyinternals.DelegateRoot):
        return {"_zeftype": "DelegateRoot"}
    elif isinstance(z, pyinternals.DelegateRelationTriple):
        return {"_zeftype": "DelegateRelationTriple", "rt": z.rt, "source": serialize_internal(z.source), "target": serialize_internal(z.target)}

    raise NotImplementedError(f"{z} not a delegate")


def serialize_zefops(k_type, ops):
    serialized_ops = []
    for op in ops:
        op_rt, op_subops = op
        assert type(op_rt) == RelationType
        op_rt = serialize_internal(op_rt)

        serialized_subops = []
        if len(op_subops) > 0 and type(op_subops[0]) == RelationType and op_subops[0] == RT.L:
            serialized_subops.append(serialize_zefops("ZefOp", (op_subops,)))
        else:
            for sub_op in op_subops:
                if type(sub_op) in list(serialization_mapping.keys()):
                    sub_op = serialize_internal(sub_op)
                serialized_subops.append(sub_op,)
        serialized_ops.append({"op": op_rt, "curried_ops": serialized_subops})

    return {"_zeftype": k_type, "el_ops": serialized_ops}

def serialize_valuetype(vt):
    # Super dodgy version just to get something off the ground for now
    return {
        "_zeftype": "ValueType",
        "type_name": vt.d["type_name"],
        "absorbed": serialize_internal(vt.d["absorbed"])
    }


def serialize_symbolicexpression(se):
    return {
        "_zeftype": "SymbolicExpression",
        "name": se.name,
        "root_node": serialize_internal(se.root_node),
        "absorbed": serialize_internal(se.absorbed)
    }


def deserialize_tuple(json_d: dict) -> tuple:
    return tuple(deserialize_internal(el) for el in json_d["items"])

def deserialize_list(l: list) -> list:
    return [deserialize_internal(el) for el in l]

def deserialize_dict(json_d):
    # return {k: deserialize(v) for k,v in json_d.items()}
    return {deserialize_internal(k): deserialize_internal(v) for k,v in json_d["items"]}

    
def deserialize_zeftypes(z) -> dict:
    if z['_zeftype'] == "ZefRef":
        g = Graph(z['guid'])
        # Note: can't use in_frame here, because if the z itself is a TX, this will not behave correctly.
        return ZefRef(g[z['uid']], g[z['tx_uid']])

    elif z['_zeftype'] == "EZefRef":
        g = Graph(z['guid'])
        return g[z['uid']] 

    # elif z['_zeftype'] == "ZefRefs" or z['_zeftype'] == "EZefRefs":
    #     g = Graph(z['guid'])
    #     if z['_zeftype'] == "ZefRefs": return ZefRefs([g[zr['uid']] | to_frame[g[z['tx_uid']]] for zr in z['value']])
    #     else: return EZefRefs([g[zr['uid']] for zr in z['value']])


    elif z['_zeftype'] in {"RT", "ET"}:
        bt_class = {"RT": RT, "ET": ET}[z['_zeftype']]
        absorbed_args = deserialize_internal(z['absorbed'])
        base = bt_class(z['value'])
        base._absorbed = absorbed_args
        return base

    elif z['_zeftype'] == "AET": 
        type_map = {
                "Int":              AET.Int,
                "Float":            AET.Float,
                "Bool":             AET.Bool,
                "String":           AET.String,
                "Enum":             AET.Enum,
                "QuantityFloat":    AET.QuantityFloat, 
                "QuantityInt":      AET.QuantityInt, 
                "Time":             AET.Time,
                "Serialized":       AET.Serialized,
        }
        absorbed_args = deserialize_internal(z['absorbed'])
        first_part,*rest = z['value'].split('.')
        out = type_map[first_part]
        for part in rest:
            out = getattr(out, part)
        if absorbed_args: out._absorbed = absorbed_args
        return out

    elif z['_zeftype'] == "Graph":
        return Graph(z['guid'])

    elif z['_zeftype'] == "Enum":
        return EN(z['enum_type'], z['enum_val'])

    elif z['_zeftype'] in {"QuantityFloat", "QuantityInt"}:
        quantity_type = {"QuantityFloat": QuantityFloat, "QuantityInt": QuantityInt}[z['_zeftype']]
        en = deserialize_internal(z['unit'])
        return quantity_type(z['value'], en)

    elif z['_zeftype'] == "Time":
        return Time(z['value']) 

    elif z['_zeftype'] in {"ZefOp", "CollectingOp", "SubscribingOp", "ForEachingOp"}:
        types = {"ZefOp": ZefOp,  "CollectingOp":CollectingOp,  "SubscribingOp":SubscribingOp, "ForEachingOp": ForEachingOp}
        if z['_zeftype'] != "ZefOp": return types[z['_zeftype']](ZefOp(deserialize_zefops(z['el_ops'])))
        return ZefOp(deserialize_zefops(z['el_ops']))

    elif z['_zeftype'] in {"LazyValue", "Awaitable"}:
        types = {"LazyValue": LazyValue,  "Awaitable":Awaitable}

        if z['_zeftype'] == "LazyValue":
            res = LazyValue(z['initial_val'])
        else:
            res = Awaitable(z['pushable'])

        res.el_ops = deserialize_internal(z['el_ops'])
        return res

    elif z['_zeftype'] == "UID":
        return uid(z['value'])

    elif z['_zeftype'] == "Image":
        import base64
        encoded_buffer = z['buffer']
        compressed_buffer = base64.b64decode(encoded_buffer)
        return Image(compressed_buffer, z['format'])

    elif z['_zeftype']  in {"Entity", "Relation", "AttributeEntity"}:
        abstract_type = {"Entity": Entity, "Relation": Relation, "AttributeEntity": AttributeEntity}[z['_zeftype']]
        uid_or_uids = "uids" if z['_zeftype'] == "Relation" else "uid"
        uid_or_uids_value = deserialize_internal(z[uid_or_uids])
        type_or_types = tuple([deserialize_internal(rae) for rae in z['type']]) if z['_zeftype'] == "Relation" else deserialize_internal(z['type'])
        absorbed_args = deserialize_internal(z['absorbed'])
        return abstract_type({'type': type_or_types, uid_or_uids: uid_or_uids_value, 'absorbed': absorbed_args})

    elif z['_zeftype'] == "ErrorType":
        return Error.__getattribute__(z['type'])(*deserialize_list(z['args']))

    elif z['_zeftype'] == "Effect":
        return deserialize_dict(z['internal_dict'])

    elif z['_zeftype'] == "FXElement":
        return FXElement(tuple(z['elements']))

    elif z['_zeftype'] == "FlatGraph":
        new_fg = FlatGraph_()
        new_fg.key_dict = deserialize_dict(z['key_dict'])
        new_fg.blobs = deserialize_tuple(z['blobs'])
        return new_fg

    elif z['_zeftype'] == "FlatRef":
        fg = deserialize_internal(z['fg'])
        return FlatRef(fg, z['idx'])

    elif z['_zeftype'] == "FlatRefs":
        fg = deserialize_internal(z['fg'])
        return FlatRefs(fg, deserialize_list(z['idxs']))

    else:
        raise NotImplementedError(f"{z['_zeftype']} isn't part of the supported deserializable zeftypes!")

def deserialize_delegate(d) -> dict:
    if d['_zeftype'] == "Delegate":
        return Delegate(d["order"], deserialize_internal(d["item"]))
    elif d['_zeftype'] == "DelegateTX":
        return pyinternals.DelegateTX()
    elif d['_zeftype'] == "DelegateRoot":
        return pyinternals.DelegateRoot()
    elif d['_zeftype'] == "DelegateRelationTriple":
        return pyinternals.DelegateRelationTriple(d["rt"], deserialize_internal(d["source"]), deserialize_internal(d["target"]))

    raise NotImplementedError(f"{d} not a serialized delegate")


def deserialize_zefops(ops):
    deserialized_ops = ()
    for op in ops:
        op_rt, op_subops = op['op'], op['curried_ops']
        assert op_rt['_zeftype'] == "RT"
        op_rt = deserialize_internal(op_rt)

        deserialized_subops = ()
        for sub_op in op_subops:
            sub_op = deserialize_internal(sub_op)
            deserialized_subops = deserialized_subops + (sub_op,)
        deserialized_ops = (*deserialized_ops , (op_rt, (*deserialized_subops,)))

    return deserialized_ops

def deserialize_valuetype(d):
    # Super dodgy version just to get something off the ground for now

    # Look for the same typename
    from . import VT
    for var in dir(VT):
        item = getattr(VT, var)
        if isinstance(item, ValueType_) and item.d["type_name"] == d["type_name"]:
            absorbed = deserialize_internal(d["absorbed"])
            return ValueType_(type_name=d["type_name"], absorbed=absorbed)
    raise Exception(f"Couldn't find a ValueType of type '{d['type_name']}'")

def deserialize_symbolicexpression(d):
    return SymbolicExpression_(
        name = d["name"],
        root_node= deserialize_internal(d["root_node"]),
        absorbed = deserialize_internal(d["absorbed"]),
    )

serialization_mapping[ZefRef] = serialize_zeftypes
# serialization_mapping[ZefRefs] = serialize_zeftypes
serialization_mapping[EZefRef] = serialize_zeftypes
# serialization_mapping[EZefRefs] = serialize_zeftypes
serialization_mapping[RelationType] = serialize_zeftypes
serialization_mapping[EntityType] = serialize_zeftypes
serialization_mapping[AttributeEntityType] = serialize_zeftypes
serialization_mapping[Graph] = serialize_zeftypes
serialization_mapping[ZefEnumValue] = serialize_zeftypes
serialization_mapping[QuantityFloat] = serialize_zeftypes
serialization_mapping[QuantityInt] = serialize_zeftypes
serialization_mapping[Time] = serialize_zeftypes
serialization_mapping[ZefOp] = serialize_zeftypes
serialization_mapping[CollectingOp] = serialize_zeftypes
serialization_mapping[SubscribingOp] = serialize_zeftypes
serialization_mapping[ForEachingOp] = serialize_zeftypes
serialization_mapping[LazyValue] = serialize_zeftypes
serialization_mapping[Awaitable] = serialize_zeftypes
serialization_mapping[BaseUID] = serialize_zeftypes
serialization_mapping[EternalUID] = serialize_zeftypes
serialization_mapping[ZefRefUID] = serialize_zeftypes
serialization_mapping[Entity] = serialize_zeftypes
serialization_mapping[Relation] = serialize_zeftypes
serialization_mapping[AttributeEntity] = serialize_zeftypes
serialization_mapping[_ErrorType] = serialize_zeftypes
serialization_mapping[Image] = serialize_zeftypes
serialization_mapping[FXElement] = serialize_zeftypes
serialization_mapping[Delegate] = serialize_delegate
serialization_mapping[FlatGraph_] = serialize_flatgraph_or_flatref
serialization_mapping[FlatRef] = serialize_flatgraph_or_flatref
serialization_mapping[FlatRefs] = serialize_flatgraph_or_flatref
serialization_mapping[pyinternals.DelegateTX] = serialize_delegate
serialization_mapping[pyinternals.DelegateRoot] = serialize_delegate
serialization_mapping[pyinternals.DelegateRelationTriple] = serialize_delegate
serialization_mapping[ValueType_] = serialize_valuetype
serialization_mapping[list] = serialize_list
serialization_mapping[tuple] = serialize_tuple
serialization_mapping[dict] = serialize_dict
serialization_mapping[SymbolicExpression_] = serialize_symbolicexpression

deserialization_mapping["dict"] = deserialize_dict
deserialization_mapping["tuple"] = deserialize_tuple
deserialization_mapping["ZefRef"] = deserialize_zeftypes
# deserialization_mapping["ZefRefs"] = deserialize_zeftypes
deserialization_mapping["EZefRef"] = deserialize_zeftypes
# deserialization_mapping["EZefRefs"] = deserialize_zeftypes
deserialization_mapping["RT"] = deserialize_zeftypes
deserialization_mapping["ET"] = deserialize_zeftypes
deserialization_mapping["AET"] = deserialize_zeftypes
deserialization_mapping["Graph"] = deserialize_zeftypes
deserialization_mapping["Enum"] = deserialize_zeftypes
deserialization_mapping["QuantityFloat"] = deserialize_zeftypes
deserialization_mapping["QuantityInt"] = deserialize_zeftypes
deserialization_mapping["Time"] = deserialize_zeftypes
deserialization_mapping["ZefOp"] = deserialize_zeftypes
deserialization_mapping["CollectingOp"] = deserialize_zeftypes
deserialization_mapping["SubscribingOp"] = deserialize_zeftypes
deserialization_mapping["ForEachingOp"] = deserialize_zeftypes
deserialization_mapping["LazyValue"] = deserialize_zeftypes
deserialization_mapping["Awaitable"] = deserialize_zeftypes
deserialization_mapping["UID"] = deserialize_zeftypes
deserialization_mapping["Entity"] = deserialize_zeftypes
deserialization_mapping["Relation"] = deserialize_zeftypes
deserialization_mapping["AttributeEntity"] = deserialize_zeftypes
deserialization_mapping["ErrorType"] = deserialize_zeftypes
deserialization_mapping["Image"] = deserialize_zeftypes
deserialization_mapping["Effect"] = deserialize_zeftypes
deserialization_mapping["FXElement"] = deserialize_zeftypes
deserialization_mapping["FlatGraph"] = deserialize_zeftypes
deserialization_mapping["FlatRef"] = deserialize_zeftypes
deserialization_mapping["FlatRefs"] = deserialize_zeftypes
deserialization_mapping["Delegate"] = deserialize_delegate
deserialization_mapping["pyinternals.DelegateTX"] = deserialize_delegate
deserialization_mapping["pyinternals.DelegateRoot"] = deserialize_delegate
deserialization_mapping["pyinternals.DelegateRelationTriple"] = deserialize_delegate
deserialization_mapping["ValueType"] = deserialize_valuetype
deserialization_mapping["SymbolicExpression"] = deserialize_symbolicexpression



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..core._ops import *
from ..core._ops import _any
from ..core import _ops, func

import types
lazy_all = [x for x in dir(_ops) if not x.startswith("_") and not isinstance(getattr(_ops, x), types.ModuleType)]

from .privileges import *
from ._config import config

from ._values import value_or, maybe_value
# from ..tools.ops import _ 

from .transactor_role import *
from ..ui import show

__all__ = [
    # "_",
    "func",
    "_any",
    "value_or",
    "maybe_value",
    "config",
    "show",
] + privileges.__all__ + lazy_all + transactor_role.__all__


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest  # pytest takes ages to run anything as soon as anything from zef is imported
from zef import *
from zef.ops import *

class MyTestCase(unittest.TestCase):

    def test_attach(self):
        g = Graph()

        z = ET.Machine | g | run
        z | fill_or_attach[RT.Weight][QuantityFloat(100.0, EN.Unit.kilogram)] | g | run
        z2 = ET.Machine | g | run
        z2 | fill_or_attach[RT.Name]["asdf"] | g | run
        z2 | fill_or_attach[RT.Status]["asdf"] | g | run
        z3 = ET.Machine | g | run
        [z3 | fill_or_attach[RT.Number][5] | collect,
         z3 | fill_or_attach[RT.Time][now()] | collect,
         z3 | fill_or_attach[RT.Fraction][0.5] | collect,
         z3 | fill_or_attach[RT.Status][EN.Status.On] | collect,
         z3 | fill_or_attach[RT.Disabled][True]
         ] | transact[g] | run

        self.assertEqual(length(z | now | out_rels[RT]), 1)
        self.assertEqual(length(z2 | now | out_rels[RT]), 2)
        self.assertEqual(length(z3 | now | out_rels[RT]), 5)

    def test_set_field(self):
        g = Graph()

        z = ET.Machine | g | run
        z | set_field[RT.Weight][QuantityFloat(100.0, EN.Unit.kilogram)] | g | run
        z2 = ET.Machine | g | run
        z2 | set_field[RT.Name]["asdf"] | g | run
        z2 | set_field[RT.Status]["asdf"] | g | run
        z3 = ET.Machine | g | run
        [z3 | set_field[RT.Number][5] | collect,
         z3 | set_field[RT.Time][now()] | collect,
         z3 | set_field[RT.Fraction][0.5] | collect,
         z3 | set_field[RT.Status][EN.Status.On] | collect,
         z3 | set_field[RT.Disabled][True]
         ] | transact[g] | run

        self.assertEqual(length(z | now | out_rels[RT]), 1)
        self.assertEqual(length(z2 | now | out_rels[RT]), 2)
        self.assertEqual(length(z3 | now | out_rels[RT]), 5)

        r = [ET.Machine["z"],
             ET.Person["y"],
             Z["z"] | set_field[RT.Supervisor][Z["y"]],
             ] | transact[g] | run
        self.assertEqual(r["z"] | Out[RT.Supervisor] | collect, r["y"])
        
        r2 = [ET.Machine["z"] | set_field[RT.Supervisor][ET.Person["y"]]] | transact[g] | run
        self.assertEqual(r2["z"] | Out[RT.Supervisor] | collect, r2["y"])

        r3 = [r["z"] | set_field[RT.Supervisor][ET.Person["new"]]] | transact[g] | run
        self.assertEqual(r["z"] | now | Outs[RT.Supervisor] | length | collect, 1)
        self.assertEqual(r["z"] | now | Out[RT.Supervisor] | collect, r3["new"])


    def test_set_field_reversed(self):
        g = Graph()

        z = ET.Machine | g | run
        z | set_field[RT.Something][42][True] | g | run

        self.assertEqual(z | now | in_rels[RT.Something] | length | collect, 1)
        self.assertEqual(z | now | In[RT.Something] | value | collect, 42)

        z | set_field[RT.Something][43][True] | g | run

        self.assertEqual(z | now | In[RT.Something] | value | collect, 43)

        z2 = ET.Machine | g | run
        z | set_field[RT.Entity][z2][True] | g | run

        self.assertEqual(z | now | In[RT.Entity] | collect, now(z2))

        z3 = ET.Machine | g | run
        z | set_field[RT.Entity][z3][True] | g | run
        self.assertEqual(z | now | In[RT.Entity] | collect, now(z3))


    def test_shorthand(self):
        g = Graph()

        z = ET.Dog | g | run
        self.assertTrue(is_a(z, ET.Dog))

        z1, [(name_rel, name),
             (breed_rel, breed)] = (z, [(RT.Name, "Rufus"),
                                        (RT.Breed, "Labrador")]) | g | run

        self.assertEqual(to_ezefref(z), to_ezefref(z1))
        self.assertTrue(is_a(name_rel, RT.Name))
        self.assertEqual(value(name), "Rufus")
        self.assertTrue(is_a(breed_rel, RT.Breed))
        self.assertEqual(value(breed), "Labrador")

        z1, none, [a,b,c] = (z, RT.Nickname, ["Dog", "Boy", "Mongrel"]) | g | run

        self.assertEqual(to_ezefref(z), to_ezefref(z1))
        self.assertEqual(none, None)
        self.assertEqual(value(a), "Dog")
        self.assertEqual(value(b), "Boy")
        self.assertEqual(value(c), "Mongrel")
        self.assertTrue(has_relation(now(z), RT.Nickname, a))
        self.assertTrue(has_relation(now(z), RT.Nickname, b))
        self.assertTrue(has_relation(now(z), RT.Nickname, c))

        [a,b], none, z1 = (["Alice", "Bob"], RT.Owner, z) | g | run

        self.assertEqual(to_ezefref(z), to_ezefref(z1))
        self.assertEqual(none, None)
        self.assertEqual(value(a), "Alice")
        self.assertEqual(value(b), "Bob")
        self.assertTrue(has_relation(a, RT.Owner, now(z)))
        self.assertTrue(has_relation(b, RT.Owner, now(z)))

    def test_shorthand_assign(self):
        g = Graph()

        ae = AET.String | g | run
        self.assertEqual(value(ae), None)

        ae | assign["word"] | g | run
        self.assertEqual(value(now(ae)), "word")

        g | now | all[AET] | single | assign["single"] | g | run
        self.assertEqual(value(now(ae)), "single")

if __name__ == '__main__':
    unittest.main()



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.



# taken from zef function
# class ToClipboard:
#     def __init__(self):
#         self.__doc__ = 'ABC'
    



#     def _make_header(self, z_zef_function: ZefRef)->str:
#         from zefdb.ops import target, uid, value, collect
#         from zefdb import tools
#         g = Graph(z_zef_function)         
#         # TODO: possibly warn if this is NOT the latest version known at the time of generating this snippet!
#         original_fct_name = z_zef_function >> RT.OriginalName | value | collect if z_zef_function | ops.has_out[RT.OriginalName] | collect else None
#         return "\n".join([
# f"# --------------------------------------------         this code snippet was generated by          --------------------------------------------",
# f"# --------------------------------------------  zefdb.to_clipboard at {str(ops.now())}  --------------------------------------------",
# #f"# origin: first created by {'-- ZefDB user --'} at ...",
# #f"# last change: by -- ZefDB user -- ...",
# # f'# z_zef_function = g["{uid(z_zef_function)}"]["{uid(z_zef_function | ops.frame)}"]',            # TODO: update this to new uid and ref frame syntax
# f'# Original function name: {original_fct_name}' if original_fct_name is not None else None,
# f"g = Graph('{uid(g)}')     # This zef function lives on this graph tagged with {g.tags}'",
# f'@func(g = g,',
# f"    label = '{value(z_zef_function >> RT.Label)}'," if z_zef_function | ops.has_out[RT.Label] | collect else None,
# f"    is_pure = {value(z_zef_function >> RT.IsPure)}," if z_zef_function | ops.has_out[RT.IsPure] | collect else None,
# *[f"    {value(z_ed>>RT.Name)} = g['{uid(z_ed | target)}'] | to_frame[g['{value(z_ed>>RT.UseTimeSlice)}']]," for z_ed in (z_zef_function > ops.L[RT.Binding] | collect) ],
# f")\n"
#         ] | ops.filter[lambda x: x is not None] | collect
#         )
    
#     def _generate_zef_fct_str(self, z_zef_function: ZefRef):  
#         from zefdb.ops import collect, value
#         z_str = z_zef_function >> RT.PythonSourceCode | collect
#         assert AET(z_str) == AET.String        
#         return self._make_header(z_zef_function) + (value(z_str))
    
#     def __call__(self, z_zef_function: Union[ZefRef, ZefRefs]):
#         if isinstance(z_zef_function, str):
#             pyperclip.copy(z_zef_function)
#             return z_zef_function
        
#         assert isinstance(z_zef_function, ZefRef) or isinstance(z_zef_function, ZefRefs)
#         z_zef_functions = [z_zef_function] if isinstance(z_zef_function, ZefRef) else z_zef_function        
#         pyperclip.copy('#%%' + '\n'*10 + ("\n"*5).join([self._generate_zef_fct_str(z_fct) for z_fct in z_zef_functions]))   # copy to clipboard
#         return z_zef_function       # allow piping through
    
#     def __ror__(self, z_zef_function):
#         return self(z_zef_function)
    

# #to be used with one ZefRef or ZefRefs to ET.ZEF_Function
# to_clipboard = ToClipboard()


from .fx_types import Effect
from ..error import Error


def clipboard_copy_to_handler(eff: dict):
    """ 
    sample effect:
    {
        'type': FX.Clipboard.CopyTo,
        'content': 'hello',
    }
    """    
    try:
        import pyperclip
    except:
        err = """
        please install pyperclip if you want to use the copy to/from
        clipboard functionality. E.g. 'pip3 install pyperclip'
        """
        raise RuntimeError(err)
    try:
        val = eff['content']
        if type(val) in {str, float, int, bool}:
            pyperclip.copy(str(val))
        else:
            raise RuntimeError(f"copying a type type(val)={type(val)} to the clipboard is not supported. Value: {val}")
        return {}
    except Exception as e:
        return Error(f'executing FX.Clipboard.CopyTo for effect {eff}:\n{repr(e)}')





def clipboard_copy_from_handler(eff: dict):
    """ 
    Get the content from the OS's clipboard. This is returned 
    in the dictionary as the value associated with the 'content'
    key.

    ---- Examples ----
    >>> {
            'type': FX.Clipboard.CopyFrom,            
        }
    """

    try:
        import pyperclip
    except:
        err = """
        please install pyperclip if you want to use the copy to/from
        clipboard functionality. E.g. 'pip3 install pyperclip'
        """
        raise RuntimeError(err)

    try:
        val = pyperclip.paste()
        return {'content': val}
    except Exception as e:
        return Error(f'executing FX.Clipboard.CopyFrom for effect {eff}:\n{repr(e)}')





# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


#=============================================================================== HTTP Server handlers =========================================================================================================
from .fx_types import Effect, FX
from ..VT import Pattern, SetOf
from uuid import uuid4
from .._ops import *
from ..error import Error

from http.server import ThreadingHTTPServer, BaseHTTPRequestHandler
import threading
from concurrent.futures import Future, TimeoutError

from ..logger import log

_effects_processes = {}

class OurHTTPServer(ThreadingHTTPServer):
    def __init__(self, *args, do_logging, **kwds):
        super().__init__(*args, **kwds)
        self.do_logging = do_logging


class Handler(BaseHTTPRequestHandler):
    def do_GET(self):
        return self.handle_generic("GET")
    def do_POST(self):
        return self.handle_generic("POST")
    def do_OPTIONS(self):
        return self.handle_generic("OPTIONS")

    def handle_generic(self, method):
        # We could override handle_one_request instead of overriding each of the
        # get_x functions, but there is error handling in handle_one_request and
        # it could change over the versions of http.server.

        try:
            future = Future()
            length = int(self.headers.get("content-length", "0"))

            body = self.rfile.read(length)

            request_id = str(uuid4())
            self.server.zef["open_requests"][request_id] = future

            splits = self.path.split("?", maxsplit=1)
            path = splits[0]
            params = splits[1] if len(splits) == 2 else ""
 
            d = {
                "request_id": request_id,
                "server_uuid": self.server.zef["server_uuid"],
                "method": method,
                "path":   path,
                "params":   params,
                "request_body":   body,
                "request_headers": dict(self.headers),      # use a plain data type
            }
            d | push[self.server.zef["stream"]] | run
            try:
                result = future.result(timeout=15.0)
            except TimeoutError:
                log.error("Timed out handling REST request")
                self.reply_with_error("Internal timeout")
                return
            finally:
                # TODO: Cleanup from open_requests here
                pass

            # TODO: This needs improving to allow sending content with non-200.
            # The only choice is when to put msg in content or in header-line
            # response.
            status,headers,msg = result
            if status != 200 and msg is not None:
                msg = msg.decode('utf-8')
                self.send_response(status, msg)
            else:
                self.send_response(status)
            for key,val in headers.items():
                self.send_header(key, val)
            self.end_headers()
            if status == 200 and msg is not None:
                self.wfile.write(msg)
        except BrokenPipeError:
            log.error("Connection aborted unexpectedly")
            pass
        except Exception as exc:
            # TODO: Place this error into a stream available to the user
            log.error("Some problem in handling HTTP request", exc_info=exc)
            self.reply_with_error("")

    def reply_with_error(self, err_msg):
        self.send_error(500, err_msg)

    def log_message(self, *args, **kwds):
        if self.server.do_logging:
            super().log_message(*args, **kwds)



def http_start_server_handler(eff: dict):
    """[summary]

    Args:
        eff : {
            "type": FX.HTTP.StartServer,
            "port": 5000, (default = 5000)
            "bind_address": "0.0.0.0", (default = "localhost")
            "pipe_into": map[..] | filter[..] | subscribe[run],
    or
            "pipe_into": map[middleware_worker[[permit_cors, custom_handle, fallback_not_found, send_response]] | subscribe[run],
            "logging": True or "Graph" or "Stream"
        }
    """
    # print(f"http_start_server called for effect: {eff}")

    server_uuid = None
    pushable_stream = None
    server = None
    thread = None
    try:
        open_requests = {}
        pushable_stream = {'type': FX.Stream.CreatePushableStream} | run
        server_uuid = str(uid(pushable_stream.stream_ezefref))
        if eff.get('pipe_into', None) is not None:
            pushable_stream | eff['pipe_into']
        port = eff.get('port', 5000)
        bind_address = eff.get('bind_address', "localhost")
        do_logging = eff.get('logging', True)

        zef_locals = {
            "open_requests": open_requests,
            "stream" : pushable_stream,
            "server_uuid": server_uuid,
            "port": port
        }

        server = OurHTTPServer((bind_address, port), Handler, do_logging=do_logging)
        zef_locals["server"] = server
        server.zef = zef_locals


        thread = threading.Thread(target=server.serve_forever, daemon=True)
        zef_locals["thread"] = thread

        _effects_processes[server_uuid] = zef_locals

        thread.start()

        log.debug(f"http_server started", uuid=server_uuid, port=port, bind_address=bind_address)
        return zef_locals
    except Exception as exc:
        raise RuntimeError(f'Error starting HTTP server: in FX.HTTP.StartServer handler: {exc}')
        


def http_stop_server_handler(eff: dict):
    d = _effects_processes[eff["server_uuid"]]
    d["server"].shutdown()
    d["server"].server_close()
    d["thread"].join()
    log.debug("http server stopped", uuid=eff["server_uuid"])
    return {}
    

def http_send_request_handler(eff: dict):
    """[summary]

    Args:
        eff (Effect): {
                "type":   FX.HTTP.Request, # Required
                "url":    "url",           # Required
                "method": "GET",           # Optional
                "data":   {},              # Optional
                "params": {},              # Optional
        }
    """
    # TODO needs more testing and improvements
    import requests
    request_url = eff['url']
    request_method = eff.get('method', 'GET')
    request_data = eff.get('data', {})
    request_params = eff.get('params', {})

    method_dict =  {'GET': requests.get, 'POST': requests.post, 'PUT': requests.put, 'DELETE': requests.delete, 'HEAD': requests.head}
    request_obj = method_dict[request_method]
    response = request_obj(url = request_url, data = request_data, params = request_params)

    return {"response_text": response.text, "response_status": response.status_code}
    
def http_send_response_handler(eff: Effect):
    d = eff
    if "server_uuid" not in d:
        # TODO: register this on ProcessLogger
        return Error(f"An FX.HTTP.SendResponse event must contain a 'server_uuid' field. This was not the case for eff={eff}")

    d = _effects_processes[eff["server_uuid"]]
    future = d["open_requests"][eff["request_id"]]

    # if "response" not in eff:
    #     print(f"Warning: FX.HTTP.SendResponse wish did not contain 'response' field. This is probably an error. Received wish: {eff}")
    msg = eff.get("response", None)
    assert msg is None or isinstance(msg, str) or isinstance(msg, bytes)
    if isinstance(msg, str):
        msg = msg.encode("utf-8")

    status = eff.get("status", 200)
    headers = eff.get("headers", {})
    import copy
    headers = copy.deepcopy(headers)

    header_names = set(k.lower() for k in headers)
    if msg is not None and status == 200:
        if "content-length" not in header_names:
            headers["Content-Length"] = str(len(msg))
        if "content-type" not in header_names:
            headers["Content-Type"] = "text/html; charset=UTF-8"

    response = (status,headers,msg)
    # log.debug("Trying to attach response to future", response=response)
    future.set_result(response)
    return {}
    
    


########################################
# * Useful middleware
#--------------------------------------

from ..zef_functions import func

# TODO: Improve this! But maybe redesign the style of the app anyway.
@func
def route(req, path, handler):
    if req["path"] == path:
        try:
            return handler(req)
        except Exception as exc:
            log.error(f"There was an exception in the handler for path {path}", exc_info=exc)
            return {
                **req,
                "response_status": 500,
            }
            
    return req

@func
def permit_cors(req):
    needed_headers = {"Access-Control-Allow-Origin": "*"}

    if req["method"] == "OPTIONS":
        # TODO: more fancy stuff here
        req = dict(**req, response_headers=needed_headers)
        req["response_headers"]["Access-Control-Allow-Methods"] = "*"
        req["response_headers"]["Access-Control-Allow-Headers"] = "*"
        req["response_headers"]["Access-Control-Max-Age"] = "86400"
        return send_response(req)

    import copy
    req = copy.deepcopy(req)

    req.setdefault("response_headers", {})
    req["response_headers"].update(needed_headers)

    return req

@func
def send_response(req):
    eff_d = {
        "type": FX.HTTP.SendResponse,
        "server_uuid": req["server_uuid"],
        "request_id": req["request_id"],
    }
    if "response_body" in req:
        eff_d["response"] = req["response_body"]
    if "response_status" in req:
        eff_d["status"] = req["response_status"]
    if "response_headers" in req:
        eff_d["headers"] = req["response_headers"]

    return eff_d

@func
def fallback_not_found(req):
    if "response_body" not in req and "response_status" not in req:
        import copy
        req = copy.deepcopy(req)
        req["response_status"] = 404
    return req


@func
def middleware_worker(req, mw):
    """Apply the functions in `mw` in order. If, at any point, None or an Effect is
returned, then immediately return the None or Effect. Otherwise, continue
applying the functions.
    
Intended to be used in the following manner:
source | map[middleware_worker[seq_of_funcs]] | run
    """

    cur = req
    for func in mw:
        cur = func(cur)

        if cur is None:
            return None
        # Not sure if this is the right way to do things
        SendResponseEffect = Pattern[{"type": SetOf(FX.HTTP.SendResponse)}]
        if is_a(cur, SendResponseEffect):
            return cur
        if isinstance(cur, dict):
            continue

        raise Exception(f"Unusual type in middleware_runner: {type(func)}")
        
    # Will only get here if the final middleware returns a dictionary
    return cur

# This doesn't work unfortunately - need to hook into the rx idea or something,
# not sure how this should look in the end.
@func
def middleware(req, mw):
    # return req | map[middleware_worker[mw]] | collect
    return req | map[middleware_worker[mw]]
    



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import *
from ...ops import *

from ...core.fx.http import send_response, permit_cors, middleware, middleware_worker, fallback_not_found, route

from .generate_api2 import generate_resolvers_fcts
from ariadne import graphql_sync

from functools import partial as P
import json

from zef.core.logger import log

def resolve_token(context, request):
    root = context["z_gql_root"]

    header = root | F.AuthHeader | collect
    aud = root | F.AuthAudience | collect
    namespace = root | Outs[RT.AuthNamespace] | single_or[None] | value_or[None] | collect
    algo = root | F.AuthAlgo | collect

    import jwt
    import base64
    headers_as_lower = {k.lower(): v for k,v in request["request_headers"].items()}
    if header.lower() not in headers_as_lower:
        return None
    token_header = headers_as_lower[header.lower()].strip()
    if token_header == '':
        return None

    parts = token_header.split()
    if len(parts) != 2 or parts[0] != "Bearer":
        # raise Exception(f"x-auth-header is of the wrong format ({token_header})")
        if context["debug_level"] >= 0:
            log.error("x-auth-header is of the wrong format", token_header=token_header)
        raise Exception("Invalid auth header")
    token = parts[1]

    if algo == "RS256":
        signing_key = context["jwk_client"].get_signing_key_from_jwt(token)
        auth_result = jwt.decode(token,
                                 signing_key.key,
                                 algorithms=[algo],
                                 audience=aud)
    elif algo == "HS256":
        psk = root | F.AuthPresharedKey | collect
        auth_result = jwt.decode(token, psk, algorithms=[algo],
                                 audience=aud)
    else:
        raise Exception("Shouldn't get here")

    if isinstance(auth_result["aud"], list):
        if aud not in auth_result["aud"]:
            raise Exception("Invalid token for wrong audience")
    elif auth_result["aud"] != aud:
        raise Exception("Invalid token for wrong audience")

    if namespace is None:
        return auth_result
    else:
        return auth_result[namespace]
            
def query(request, context):
    root = context["z_gql_root"]
    if root | has_out[RT.AuthHeader] | collect:
        try:
            auth_context = resolve_token(context, request)
            if auth_context is None and not (root | Outs[RT.AuthPublic] | single_or[None] | value_or[True] | collect):
                raise Exception("No auth and public is False.")
        except Exception as exc:
            if context["debug_level"] >= 0:
                log.error("Auth failed", exc_info=exc)
            return {
                "response_body": "Auth failed",
                "response_status": 400,
                **request
            }

        if context["debug_level"] >= 4:
            log.debug("DEBUG 4: auth_context", auth_context=auth_context)
    else:
        auth_context = None

    if request["method"] == "GET" or request["request_body"].strip() == "":
        # This is to give some kind of response instead of a failure. Helps with
        # health checks.
        return {
            "response_body": "Server waiting for GraphQL requests",
            **request,
        }

    q = json.loads(request["request_body"])

    if context["debug_level"] >= 3:
        log.debug("DEBUG 3: incoming query", q=q)

    # We pass in the graph as a fixed slice, so that the queries can be done
    # consistently.
    success,data = graphql_sync(
        context["ari_schema"],
        q,
        context_value={"gs": now(context["g_data"]),
                       "auth": auth_context,
                       "debug_level": context["debug_level"]},
    )
    if context["debug_level"] >= 0:
        if not success or "errors" in data:
            log.error("Failure in GQL query.", data=data, q=q, auth_context=auth_context)

    response = json.dumps(data)
    if context["debug_level"] >= 3:
        log.debug("DEBUG 3: response", response=response)

    return {
        "response_body": response,
        **request
    }

def start_server(z_gql_root,
                 g_data,
                 *,
                 port=443,
                 bind_address="0.0.0.0",
                 logging=True,
                 debug_level=0,
                 ):

    gql_dict = generate_resolvers_fcts(z_gql_root)
    from .. import make_graphql_api
    ari_schema = make_graphql_api(gql_dict)

    from logging import getLogger
    if debug_level < 4:
        logger = getLogger("ariadne").disabled = True
        log.debug("Disabled ariadne logging")
    
    context = {
        "z_gql_root": z_gql_root,
        "g_data": g_data,
        "ari_schema": ari_schema,
        "debug_level": debug_level,
    }

    if z_gql_root | has_out[RT.AuthJWKURL] | collect:
        url = z_gql_root | F.AuthJWKURL | collect

        from jwt import PyJWKClient
        context["jwk_client"] = PyJWKClient(url)

    additional_routes = create_additional_routes(z_gql_root | Outs[RT.Route] | collect, context)

    http_r = {
        'type': FX.HTTP.StartServer,
        'port': port,
        'pipe_into': (map[middleware_worker[[permit_cors,
                                             route["/"][insert_in[["response_body"]]["Healthy"]],
                                             route["/gql"][P(query, context=context)],
                                             *additional_routes,
                                             fallback_not_found,
                                             send_response]]]
                      | subscribe[run]),
        'logging': logging,
        'bind_address': bind_address,
    } | run
    if is_a(http_r, Error):
        raise Exception("Error in creating server") from http_r.args[0]

    return http_r["server_uuid"]


def create_additional_route(z_route, context):
    s_route = z_route | F.Route | collect
    hook = func[z_route | F.Hook | collect]

    curried_hook = P(hook, context=context)

    return route[s_route][curried_hook]

def create_additional_routes(z_routes, context):
    return z_routes | map[P(create_additional_route, context=context)] | collect


        


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
SymbolicVariables are a special kind of SymbolicExpression. They are a primitive building block.
Which variable they are is simply stored as a string, which makes them values without an identity
other than this string.

There are also other kinds of SymbolicExpressions: they can be composed to form composite expressions.
Internally, the structure of the expression is stored as a FlatGraph (this is a more suitable
data structure than a tree).

Generally, SymbolicExpressions follow strict value semantics though. If a new expression is expressed
in terms of another SymbolicExpression, the original expression is never mutated.

It is generally up to the outer scope to decide what to do with a SymbolicExpression. They are just
a value, potentially a complicated one.

They are useful in Queries, expressing various interconnected constraints, truth maintenance systems,
logic programs that are built up incrementally.

A very similar structure is used for ReactiveVariables. These are a different type and have a different
use case though.




----------------------------- Basic Usage -----------------------------------
>>> x = SV('x')
>>> y = x*x + 42     # y is a composite SymbolicExpression

Many operators are supported


----------------------------- Use Cases -----------------------------------
# z is a SE: serve as the body of a Zef Lambda expression.
Lambda[z*z + 42]

p, a = SVs('a, p')
Query[
    {
        p: ET.Person,
        a: ET.Dog | ET.Cat,
    },
    RP(p, RT.Owns, a),
]


Query[
    z,
    z[1]*z[1] + z[2]*z[2] == z[3]*z[3]
]

or
Query[
    (x1, x2, x3),               # do we want this? Not entirely consistent with the Dict notation.
    x1*x1 + x2*x2 == x3*x3,
]


----------------------------- Sample Code -----------------------------------
>>> from zef.core.symbolic_expression import SymbolicExpression_, SV, unwrap_vars
>>> 
>>> x1 = SV('x1')
>>> x2 = SV('x2')
>>> expr = (x1 + 5)**x2
>>> expr.root_node.fg | unwrap_vars | graphviz | collect   # temporary hack: shows graphviz representation of the expression

----------------------------- Open Questions -----------------------------------
- What should "List[x2]" do? Simply Absorb the SymbolicExpression or should this itself be a SymbolicExpression?
- What is the grammar for SEs interacting with ReactiveVariables?
- if we expose `z`, should z['a'] be identical to SV('a')?

----------------------------- TODOs -----------------------------------
- add SymbolicExpression to VTs and implement in is_a
- serialization
- overload some unary operators: e.g. __pos__, __neg__, ...?
- 41 | add[1] | x | multiply[2]          # advanced use case: should this also be a SE? SEs can themselves be used to do logic with ZefOps.

"""


from .VT import FlatGraph, Pattern, Any, SetOf
from ._ops import match, collect, insert, split, get, filter, map, Z
from .zef_functions import func
from .internals import ET, RT, AET
from .flat_graph import Val
from .z_field import ZField_

def merge_flatgraphs(g1, g2) -> FlatGraph:
    """
    Given two graphs, make a new FlatGraph that contains both.

    """
    def shift(el, d):
        """
        depending on the element type in the FlatGraph blobs, 
        we may have to perform different types of shifts.
        """
        return el | match[
         (Pattern[Any, AET], lambda el: (el[0]+d, el[1], [n+d if n>=0 else n-d for n in el[2]], el[3], el[4])),
         (Pattern[Any, RT], lambda el: (el[0]+d, el[1], [n+d if n>=0 else n-d for n in el[2]], el[3], el[4]+d, el[5]+d)),
         (Any, lambda el: (el[0]+d, el[1], [n+d if n>=0 else n-d for n in el[2]], el[3],)),
        ] | collect
    

    g = FlatGraph()
    delta_n = len(g1.blobs)    # how much to shift all indexes of g2
    g.blobs = (
        *g1.blobs,
        *(shift(el, delta_n) for el in g2.blobs),
    )
    g.key_dict = {**g1.key_dict, **{k: v+delta_n for k,v in g2.key_dict.items()}}
    return g




class SymbolicExpression_:
    """Symbolic Variable class"""
    def __init__(self, name=None, root_node=None, absorbed=None):
        self.name = name
        self.root_node = root_node
        self.absorbed = absorbed

    def __repr__(self):
        if self.name is not None:
            # if the name is specified: it is a SV only
            return f"v.{self.name}" if self.absorbed==None else f"v.{self.name}['{self.absorbed}']"
            # return f"SV('{self.name}')" if self.absorbed==None else f"SV('{self.name}')['{self.absorbed}']"
            # return f"{self.name}" if self.absorbed==None else f"{self.name}['{self.absorbed}']"
            
        else:
            return "Composed SymbolicExpression (todo: expression output)"

    def __hash__(self):
        return (hash(self.name)+435677842)^hash(self.absorbed)^(hash(self.root_node)+3424242)


    def __getitem__(self, k):
        if self.root_node is not None:
            raise RuntimeError("a composite SymbolicExpression cannot absorb a value")
        return SymbolicExpression_(name=self.name, absorbed=k)

    def __getattr__(self, name):
        return ZField_(name)

    def __add__(self, other):
        return compose_se(ET.Add, self, other)

    def __radd__(self, other):
        return compose_se(ET.Add, other, self)

    def __sub__(self, other):
        return compose_se(ET.Subtract, self, other)

    def __rsub__(self, other):
        return compose_se(ET.Subtract, other, self)

    def __mul__(self, other):
        return compose_se(ET.Multiply, self, other)

    def __rmul__(self, other):
        return compose_se(ET.Multiply, other, self)

    def __truediv__(self, other):
        return compose_se(ET.Divide, self, other)

    def __rtruediv__(self, other):
        return compose_se(ET.Divide, other, self)

    def __eq__(self, other):
        return compose_se(ET.Equals, self, other)

    def __req__(self, other):
        return compose_se(ET.Equals, other, self)

    def __ne__(self, other):
        return compose_se(ET.NotEquals, self, other)

    def __rne__(self, other):
        return compose_se(ET.NotEquals, other, self)

    def __gt__(self, other):
        return compose_se(ET.GreaterThan, self, other)

    def __rgt__(self, other):
        return compose_se(ET.GreaterThan, other, self)

    def __ge__(self, other):
        return compose_se(ET.GreaterOrEqualThan, self, other)

    def __rge__(self, other):
        return compose_se(ET.GreaterOrEqualThan, other, self)

    def __lt__(self, other):
        return compose_se(ET.LessThan, self, other)

    def __rlt__(self, other):
        return compose_se(ET.LessThan, other, self)

    def __le__(self, other):
        return compose_se(ET.LessOrEqualThan, self, other)

    def __rle__(self, other):
        return compose_se(ET.LessOrEqualThan, other, self)

    def __mod__(self, other):
        return compose_se(ET.Modulus, self, other)

    def __rmod__(self, other):
        return compose_se(ET.Modulus, other, self)

    def __or__(self, other):
        return compose_se(ET.Or, self, other)

    def __ror__(self, other):
        return compose_se(ET.Or, other, self)

    def __and__(self, other):
        return compose_se(ET.And, self, other)

    def __rand__(self, other):
        return compose_se(ET.And, other, self)

    def __xor__(self, other):
        return compose_se(ET.Xor, self, other)

    def __rxor__(self, other):
        return compose_se(ET.Xor, other, self)

    def __pow__(self, other):
        return compose_se(ET.Power, self, other)

    def __rpow__(self, other):
        return compose_se(ET.Power, other, self)



def SV(name: str):
    """acts like a constructor. SymbolicVariable is the `primitive`, 
    atomic building block  of any symbolic expression. The latter
    can be a larger composed form."""
    if not isinstance(name, str):
        raise TypeError("name must be a string")    
    return SymbolicExpression_(name=name)


def SVs(names: str):
    """
    initialize multiple SVs together:
    >>> x1, x2, *_ = SVs('x1, x2, x3, x4')
    Space or comma indicates a new variable.
    """
    if not isinstance(names, str):
        raise TypeError("name must be a string")    

    return (names 
        | split[lambda c: c==' ' or c==','] 
        | filter[~SetOf['',' ', ',']] 
        | map[SV]
        | collect
        )



def compose_se(op_type, arg1, arg2):
    """
    either arg1 or arg2 will be a SymbolicExpression.
    Otherwise the magic method would not forward to this function.
    """
    fg = FlatGraph()

    is_composite_se = lambda x: isinstance(x, SymbolicExpression_) and x.name == None
    arg1_composite_se = is_composite_se(arg1)
    arg2_composite_se = is_composite_se(arg2)

    # neither side is a composite SE: start with fresh graph
    if (not arg1_composite_se) and (not arg2_composite_se):
        res = (fg 
            | insert[op_type['root'], RT.Arg1, Val(arg1)]
            | insert[Z['root'], RT.Arg2, Val(arg2)]
            | get['root']
            | collect
        )
        return SymbolicExpression_(root_node=res)
    
    # both sides are composite SEs: merge
    elif arg1_composite_se and arg2_composite_se:
        g1 = arg1.root_node.fg
        g2 = arg2.root_node.fg
        g_merged = merge_flatgraphs(g1, g2)
        # some manual hacking here. g2's blob list is appended to g1's blob list.
        g_merged.key_dict['_arg1'] = arg1.root_node.fg.key_dict['root']
        g_merged.key_dict['_arg2'] = arg2.root_node.fg.key_dict['root'] + len(arg1.root_node.fg.blobs)
        res = (g_merged
            | insert[op_type['root'], RT.Arg1, Z['_arg1']]
            | insert[Z['root'], RT.Arg2, Z['_arg2']]
            | get['root']
            | collect
        )
        return SymbolicExpression_(root_node=res)
        

    elif arg1_composite_se:
        res = (arg1.root_node.fg
            | insert[op_type['root'], RT.Arg1, arg1.root_node]
            | insert[Z['root'], RT.Arg2, Val(arg2)]
            | get['root']
            | collect
        )
        return SymbolicExpression_(root_node=res)
        
    elif arg2_composite_se:
        res = (arg2.root_node.fg
            | insert[op_type['root'], RT.Arg2, arg2.root_node]
            | insert[Z['root'], RT.Arg1, Val(arg1)]
            | get['root']
            | collect
        )
        return SymbolicExpression_(root_node=res)
        
    raise RuntimeError('We should not be here')



@func
def unwrap_vars_hack(fg):
    """
    unwrap variables from a flatgraph.
    """
    def unwrap(x):
        try:
            return (*x[:3], x[3].name)
        except:
            return x
    fg2 = FlatGraph()
    fg2.key_dict = fg.key_dict
    fg2.blobs = tuple((unwrap(x) for x in fg.blobs))
    return fg2



class ZefVariable_():
    """
    A helper class for a shorthand way to 
    construct a variable called "x2": v.x2
    """
    def __getattr__(self, name):
        return SV(name)


v = ZefVariable_()





# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from cogapp import Cog

import sys
search_path = sys.argv[1]
output_path = sys.argv[2]
tokens_path = sys.argv[3]

def find_files_of_type(path, filename_endings, directories_to_exclude={}):
    # Returns paths relative to the input path
    for dirpath,dirnames,filenames in os.walk(path):
        for filename in filenames:
            thispath = os.path.join(dirpath, filename)
            if os.path.splitext(thispath)[1] in filename_endings:
                yield os.path.relpath(thispath, path)

        for dirname in directories_to_exclude:
            while dirname in dirnames:
                dirnames.remove(dirname)

cog = Cog()
cog.options.bReplace = False
cog.options.bDeleteCode = True
cog.options.sPrologue = f"""
import cog
import json
import os

et_filename = os.path.join("{tokens_path}", "zeftypes_ET.json")
with open(et_filename) as F:
    et = json.loads(F.read())

rt_filename = os.path.join("{tokens_path}", "zeftypes_RT.json")
with open(rt_filename) as F:
    rt = json.loads(F.read())

kw_filename = os.path.join("{tokens_path}", "zeftypes_KW.json")
with open(kw_filename) as F:
    kw = json.loads(F.read())

en_filename = os.path.join("{tokens_path}", "zeftypes_EN.json")
with open(en_filename) as F:
    en = json.loads(F.read())

def enum_type(x):
    return x.split('.')[0]
def enum_val(x):
    return x.split('.')[1]
"""


#cog.options.verbosity = 0
for filename in find_files_of_type(path=search_path, filename_endings={'.cog'}):
    try:
        true_output = os.path.join(output_path, filename[:-len(".cog")] + ".gen")
        os.makedirs(os.path.dirname(true_output), exist_ok=True)
        cog.options.sOutputName = true_output + ".tmp"
        cog.processOneFile(os.path.join(search_path, filename))
        if not os.path.exists(true_output) or open(true_output + ".tmp").read() != open(true_output).read():
            print(filename, " changed")
            os.rename(true_output + ".tmp", true_output)
        else:
            os.unlink(true_output + ".tmp")
    except Exception as exc:
        print(f'An exception was raised when processing file "{filename}": {exc}')
        # Need this to fail for cmake to not continue on without a care.
        raise



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import builtins
from typing import Generator, Iterable, Iterator


from rx import operators as rxops
from ..VT.value_type import ValueType_

# This is the only submodule that is allowed to do this. It can assume that everything else has been made available so that it functions as a "user" of the core module.
from .. import *
from .._ops import *
from ..logger import log
from ...pyzef import zefops as pyzefops, main as pymain
from .. import internals
from typing import Generator, Iterable, Iterator


#-----------------------------FlatGraph Implementations-----------------------------------
def fg_insert_imp(fg, new_el):
    from ..graph_delta import map_scalar_to_aet_type
    from ...pyzef.internals import DelegateRelationTriple

    assert is_a(fg, FlatGraph)
    new_fg = FlatGraph()
    new_blobs, new_key_dict = [*fg.blobs], {**fg.key_dict}

    def idx_generator(n):
        def next_idx():
            nonlocal n
            n = n + 1
            return n
        return next_idx

    next_idx = idx_generator(length(fg.blobs) - 1)

    def inner_zefop_type(zefop, rt):
        return peel(zefop)[0][0] == rt

    def construct_abstract_rae_and_return_idx(rae_type, rae_uid):
        if isinstance(rae_type, RelationType):
            assert rae_uid in new_key_dict, "Can't construct an Abstract Relation!"
            return new_key_dict[rae_uid]
        else:
            rae_class = AttributeEntity if type(rae_type) == AttributeEntityType else Entity
            return common_logic(rae_class({"type": rae_type, "uid": rae_uid}))

    def common_logic(new_el):
        nonlocal new_blobs, new_key_dict

        if isinstance(new_el, EntityType):
            idx = next_idx()
            internal_id = new_el | absorbed | attempt[single][None] | collect
            new_el = new_el | without_absorbed | collect
            if internal_id: new_key_dict[internal_id] = idx
            new_blobs.append((idx, new_el, [], None))

        elif isinstance(new_el, AttributeEntityType):
            idx = next_idx()
            internal_id = new_el | absorbed | attempt[single][None] | collect
            new_el = new_el | without_absorbed | collect
            if internal_id: new_key_dict[internal_id] = idx
            new_blobs.append((idx, new_el, [], None, None))

        elif isinstance(new_el, Entity):
            node_type, node_uid = new_el.d['type'], new_el.d['uid']
            if node_uid not in new_key_dict:
                idx = next_idx()
                new_blobs.append((idx, node_type, [], node_uid))
                new_key_dict[node_uid] = idx
            idx = new_key_dict[node_uid]

        elif isinstance(new_el, AttributeEntity):
            node_type, node_uid = new_el.d['type'], new_el.d['uid']
            if node_uid not in new_key_dict:
                idx = next_idx()
                new_blobs.append((idx, node_type, [], node_uid, None))
                new_key_dict[node_uid] = idx
            idx = new_key_dict[node_uid]

        elif isinstance(new_el, ZefOp) and inner_zefop_type(new_el, RT.Instantiated):
            raise ValueError("!!!!SHOULD NO LONGER ARRIVE HERE!!!!")
        
        elif isinstance(new_el, ZefOp) and inner_zefop_type(new_el, RT.Terminate):
            to_be_removed = LazyValue(new_el) | absorbed | attempt[first][None] | collect
            idx = None
            if to_be_removed:
                try:
                    new_fg = FlatGraph()
                    new_fg.key_dict = new_key_dict
                    new_fg.blobs = (*new_blobs,)
                    new_fg = fg_remove_imp(new_fg, to_be_removed)
                    new_blobs, new_key_dict = [*new_fg.blobs], {**new_fg.key_dict}
                except KeyError:
                    pass
                except:
                    raise Exception(f"An exception happened while trying to perform {new_el} on {fg}")

        elif isinstance(new_el, ZefOp) and inner_zefop_type(new_el, RT.Assign):
            internals = LazyValue(new_el) | absorbed  | collect
            assert len(internals) == 2, f"assign should have both the entity and the value to be assigned got {internals} instead"
            ref, val = internals
            idx = common_logic(ref <= val)


        elif isinstance(new_el, ZefOp) and inner_zefop_type(new_el, RT.Z):
            key = peel(new_el)| first | second | first | collect
            if key not in new_key_dict and not isinstance(key, int): raise KeyError(f"{key} doesn't exist in internally known ids!")
            idx = new_key_dict.get(key, key)

        # i.e: z4 <= 42 ; AET.String <= "42" ; AET.String['z1'] <= 42 ; Z['n1'] <= 42
        elif isinstance(new_el, LazyValue) and length(peel(new_el)) == 2:
            first_op = peel(new_el)[0]
            if isinstance(first_op, AttributeEntityType):
                    internal_id = first_op | absorbed | attempt[single][None] | collect
                    aet_maybe = first_op | without_absorbed | collect
                    assert isinstance(aet_maybe, AttributeEntityType), f"{new_el} should be of type AET"
                    aet_value = peel(peel(new_el)[1])[0][1][0]
                    idx = next_idx()
                    new_blobs.append((idx, aet_maybe, [], None, aet_value))
                    if internal_id: new_key_dict[internal_id] = idx
            elif isinstance(first_op, ZefOp):
                if inner_zefop_type(first_op, RT.Z):
                    key = peel(first_op)[0][1][0]
                    aet_value = peel(peel(new_el)[1])[0][1][0]
                    if key not in new_key_dict and not isinstance(key, int): raise KeyError(f"{key} doesn't exist in internally known ids!")
                    idx = new_key_dict.get(key, key)
                    assert isinstance(new_blobs[idx][1], AttributeEntityType), f"This key must refer to an AET found {new_blobs[idx][1]}"
                    new_blobs[idx] = (*new_blobs[idx][:4], aet_value)
                else:
                    raise ValueError(f"Expected a Z['n1'] <= 42 got {new_el} instead!")
            elif isinstance(first_op, ZefRef) or isinstance(first_op, EZefRef):
                idx = common_logic(first_op)
                assert isinstance(new_blobs[idx][1], AttributeEntityType), f"This key must refer to an AET found {new_blobs[idx][1]}"
                aet_value = peel(peel(new_el)[1])[0][1][0]
                new_blobs[idx] = (*new_blobs[idx][:4], aet_value)
            else:
                raise ValueError(f"Expected a z <= 42 or AET.String <= 42 got {new_el} instead!")

        elif isinstance(new_el, Val):
            new_el = new_el.arg
            hash_vn = value_hash(new_el)
            if hash_vn not in new_key_dict:
                idx = next_idx()
                new_key_dict[hash_vn] = idx
                new_blobs.append((idx, "BT.ValueNode", [], new_el))  # TODO Don't treat as str once added to Zef types
            idx = new_key_dict[hash_vn]

        elif isinstance(new_el, FlatRef):
            if new_el.fg == fg:
                idx = new_el.idx
            else:
                # If the flatgraphs are different then merge the FlatGraph in and return the
                # new index of the blob originally in the other FlatGraph
                idx = fr_merge_and_retrieve_idx(new_blobs, new_key_dict,next_idx, new_el)

        elif type(new_el) in list(map_scalar_to_aet_type.keys()): 
            aet = map_scalar_to_aet_type[type(new_el)](new_el)
            idx = next_idx()
            new_blobs.append((idx, aet, [], None, new_el))

        elif type(new_el) in {ZefRef, EZefRef}:
            idx = common_logic(origin_rae(new_el))
            if isinstance(new_blobs[idx][1], AttributeEntityType) and isinstance(new_el, ZefRef):
                new_blobs[idx] = (*new_blobs[idx][:4], value(new_el))
        elif isinstance(new_el, Delegate):
            if isinstance(new_el.item, DelegateRelationTriple):
                if new_el in new_key_dict:
                    idx = new_key_dict[new_el]
                else:
                    src, rt, trgt = source(new_el), new_el.item.rt, target(new_el)
                    src_idx = common_logic(src)
                    trgt_idx = common_logic(trgt)
                    idx = next_idx()

                    new_blobs.append((idx, new_el, [], None, src_idx, trgt_idx))
                    if idx not in new_blobs[src_idx][2]: new_blobs[src_idx][2].append(idx)
                    if idx not in new_blobs[trgt_idx][2]: new_blobs[trgt_idx][2].append(-idx)
                    new_key_dict[new_el] = idx
            else:
                internal_id = new_el | absorbed | attempt[single][None] | collect
                new_el = new_el | without_absorbed | collect
                if new_el in new_key_dict:
                    idx = new_key_dict[new_el]
                else:
                    idx = next_idx()
                    if internal_id: new_key_dict[internal_id] = idx
                    new_key_dict[new_el] = idx
                    new_blobs.append((idx, new_el, [], None))
        else:
            idx = None
        return idx
    
    def _insert_dict(new_el):
        ent, sub_d = list(new_el.items()) | single | collect
        ent_idx = common_logic(ent)
        for k,v in sub_d.items():
            if isinstance(v, dict):
               target_idx =  _insert_dict(v)
               _insert_single((Z[ent_idx], k, Z[target_idx]))
            else:
                _insert_single((Z[ent_idx], k, v))
        return ent_idx

    def _insert_single(new_el):
        if type(new_el) in {EntityType, AttributeEntityType, Entity, AttributeEntity, ZefOp, LazyValue,ZefRef, EZefRef, *list(map_scalar_to_aet_type.keys()), Val, Delegate}:
            common_logic(new_el)
        elif isinstance(new_el, tuple) and len(new_el) == 3:
            src, rt, trgt = new_el
            src_idx = common_logic(src)
            trgt_idx = common_logic(trgt)
            assert isinstance(src_idx, int) and isinstance(trgt_idx, int), "Couldn't find/create src or target!"
            assert type(rt) in {RelationType, ZefOp}, "Tuples must have Relation as second item."
            idx = next_idx()

            # Case of RT.A['a']
            if isinstance(rt, RelationType):
                internal_id = LazyValue(rt) | absorbed | attempt[single][None] | collect
                rt = LazyValue(rt) | without_absorbed | collect
                if internal_id: new_key_dict[internal_id] = idx
            # Case of Z['a']
            elif isinstance(rt, ZefOp) and inner_zefop_type(rt, RT.Z): 
                raise ValueError(f"Cannot reference an internal element to be used as a Relation. {rt}")

            new_blobs.append((idx, rt, [], None, src_idx, trgt_idx))
            if idx not in new_blobs[src_idx][2]: new_blobs[src_idx][2].append(idx)
            if idx not in new_blobs[trgt_idx][2]: new_blobs[trgt_idx][2].append(-idx)
        elif isinstance(new_el, Relation):
            src, rt, trgt = new_el.d['type']
            src_uid, rt_uid, trgt_uid = new_el.d['uids']

            if type(src) == RelationType and src_uid not in new_key_dict: raise ValueError("Source of an abstract Relation can't be a Relation that wasn't inserted before!")
            if type(trgt) == RelationType and trgt_uid not in new_key_dict: raise ValueError("Target of an abstract Relation can't be a Relation that wasn't inserted before!")
            src_idx = construct_abstract_rae_and_return_idx(src, src_uid)
            trgt_idx = construct_abstract_rae_and_return_idx(trgt, trgt_uid)
            idx = next_idx()
            new_blobs.append((idx, rt, [], rt_uid, src_idx, trgt_idx))
            new_key_dict[rt_uid] = idx
            if idx not in new_blobs[src_idx][2]: new_blobs[src_idx][2].append(idx)
            if idx not in new_blobs[trgt_idx][2]: new_blobs[trgt_idx][2].append(-idx)
        elif isinstance(new_el, dict): 
                _insert_dict(new_el)
        else: 
            raise NotImplementedError(f"Insert not implemented for {type(new_el)}.\nnew_el={new_el}")
        
    if isinstance(new_el, list): 
        def sorting_key(el):
            if isinstance(el, dict): return 1
            elif isinstance(el, Relation): return 2
            elif isinstance(el, tuple) and len(el) == 3: 
                is_z = lambda el: isinstance(el, ZefOp) and inner_zefop_type(el, RT.Z)
                has_internal_id = lambda rt: isinstance(rt, RelationType) and (LazyValue(rt) | absorbed | attempt[single][None] | collect) != None
                return 3 + sum([is_z(el) for el in el]) - has_internal_id(el[1])
            return 0
        new_el.sort(key=sorting_key)
        [_insert_single(el) for el in new_el]
    elif isinstance(new_el, dict): 
        _insert_dict(new_el)
    else: 
        _insert_single(new_el)
        
    new_fg.key_dict = new_key_dict
    new_fg.blobs = (*new_blobs,)
    return new_fg

def fr_merge_and_retrieve_idx(blobs, k_dict, next_idx, fr):
    fr_idx = fr.idx
    fg2 = fr.fg

    idx_key_2 = {i:k for k,i in fg2.key_dict.items()}
    old_to_new = {}

    def retrieve_or_insert_blob(new_b):
        old_idx = new_b[0]
        key = idx_key_2.get(new_b[0], None)

        if old_idx in old_to_new:
            idx = old_to_new[old_idx]
            new_b = blobs[idx]
        elif (new_b[1] == 'BT.ValueNode' or is_a(new_b[3], uid)) and key in k_dict:
            new_b = blobs[k_dict[key]]
            idx = new_b[0]
        else:
            idx = next_idx()
            new_b = (idx, new_b[1], [], *new_b[3:])
            blobs.append(new_b)
        old_to_new[old_idx] = idx
        return new_b

    for b in fg2.blobs | filter[lambda b: isinstance(b[1], RelationType)] | sort[lambda b: -len(b[2])]:
        src_b, trgt_b = fg2.blobs[b[4]], fg2.blobs[b[5]]
        src_b  = retrieve_or_insert_blob(src_b)
        trgt_b = retrieve_or_insert_blob(trgt_b)
        
        idx = next_idx()
        rt_b = (idx, b[1], [], None, src_b[0], trgt_b[0])
        src_b[2].append(idx)
        trgt_b[2].append(-idx)
        blobs.append(rt_b)
        old_to_new[b[0]] = idx

            
    for b in fg2.blobs | filter[lambda b: not isinstance(b[1], RelationType)]:
        if b[0] not in old_to_new: retrieve_or_insert_blob(b)

    # If there is no overlab with the main k_dict, we insert
    # the old key with the new index
    for k,v in fg2.key_dict.items():
        if k not in k_dict: k_dict[k] = old_to_new[v]

    return old_to_new[fr_idx]

def fg_get_imp(fg, key):
    kdict = fg.key_dict
    if type(key) in {ZefRef, EZefRef} and origin_uid(key) in kdict: return FlatRef(fg, kdict[origin_uid(key)])
    elif type(key) in {Entity, AttributeEntity} and key.d['uid'] in kdict: return FlatRef(fg, kdict[key.d['uid']])
    elif isinstance(key, Relation) and key.d['uids'][1] in kdict:return FlatRef(fg, kdict[key.d['uids'][1]])
    elif isinstance(key, Val) and  value_hash(key.arg) in kdict: return FlatRef(fg, kdict[value_hash(key.arg)])
    elif key in kdict: return FlatRef(fg, kdict[key])
    else: raise KeyError(f"{key} isn't found in this FlatGraph!")


def fg_remove_imp(fg, key):
    error = KeyError(f"{key} isn't found in this FlatGraph!")
    kdict = fg.key_dict
    if type(key) in {ZefRef, EZefRef}:  
        if origin_uid(key) not in kdict: raise error
        idx = kdict[origin_uid(key)]
        key = origin_uid(key)
    elif type(key) in {Entity, AttributeEntity}: 
        if key.d['uid'] not in kdict: raise error
        idx = kdict[key.d['uid']]
        key = key.d['uid']
    elif isinstance(key, Relation):
        if key.d['uids'][1] not in kdict: raise error
        idx = kdict[key.d['uids'][1]]
        key = key.d['uids'][1]
    elif isinstance(key, Val):
        if value_hash(key.arg) not in kdict: raise error
        idx = kdict[value_hash(key.arg)]
        key = value_hash(key.arg)
    elif key in kdict:  
        idx = kdict[key]
    else: raise error

    idx_key = {idx:key for key,idx in kdict.items()}
    kdict   = {**fg.key_dict}
    blobs   = [*fg.blobs]

    def remove_blob(idx, key = None):
        blob  = blobs[idx]
        if blob:
            blob_type = blob[1]
            ins_outs  = blob[2]
            blobs[idx] = None
            if not key: 
                key = idx_key.get(idx, None)
                if key : del(kdict[key])
            else: del(kdict[key])
            if type(blob_type) == RelationType:
                src_idx, trgt_idx = blob[4:]
                if blobs[src_idx] and idx in blobs[src_idx][2]: blobs[src_idx][2].remove(idx)
                if blobs[trgt_idx] and -idx in blobs[trgt_idx][2]: blobs[trgt_idx][2].remove(-idx)
            ins_outs | map[abs] | for_each[remove_blob]
    remove_blob(idx, key)

    new_fg = FlatGraph()
    new_fg.blobs = blobs
    new_fg.key_dict = kdict
    return new_fg


def flatgraph_to_commands(fg):
    return_elements = set()
    idx_key = {idx:key for key,idx in fg.key_dict.items()}
    def dispatch_on_blob(b, for_rt=False):
        idx = b[0]
        if isinstance(b[1], EntityType):
            if idx in idx_key:
                key = idx_key[idx]
                if is_a(key, uid): return Entity({"type": b[1], "uid": key})
                else:
                    if for_rt: return Z[key]
                    return b[1][key]
            else:
                if b[1][idx] in return_elements: return Z[idx]
                return  b[1][idx] 
        elif isinstance(b[1], AttributeEntityType):
            if idx in idx_key:
                key = idx_key[idx]
                if is_a(key, uid):
                    if b[-1] != None: return AttributeEntity({"type": b[1], "uid": key}) <= b[-1]
                    else:     return AttributeEntity({"type": b[1], "uid": key})
                else:
                    if for_rt: return Z[key]
                    if b[-1] != None: return b[1][key] <= b[-1]
                    else:     return b[1][key]
            else:
                if b[-1] != None: 
                    if (b[1][idx] <= b[-1]) in return_elements: return Z[idx] <= b[-1]
                    return b[1][idx] <= b[-1]
                else:     
                    if (b[1][idx]) in return_elements: return Z[idx]
                    return b[1][idx]
        elif isinstance(b[1], RelationType):
            if idx in idx_key: 
                key = idx_key[idx]
                if for_rt: return Z[key]
                if is_a(key, uid):
                    return Relation({"type": (fg.blobs[b[4]][1],b[1], fg.blobs[b[5]][1]), "uids": (idx_key[b[4]], key, idx_key[b[5]])})
            if for_rt: return Z[idx]
            src_blb  = dispatch_on_blob(fg.blobs[b[4]], True)
            trgt_blb = dispatch_on_blob(fg.blobs[b[5]], True)
            if b[0] in idx_key: base = b[1][idx_key[b[0]]]
            else: base = b[1][idx]
            return (src_blb, base, trgt_blb)

        elif isinstance(b[1], str) and b[1][:2] == "BT":
            if for_rt:
                return Z[value_hash(b[-1])]
            else:
                from ..graph_delta import map_scalar_to_aet_type
                if type(b[-1]) in map_scalar_to_aet_type:
                    aet = map_scalar_to_aet_type[type(b[-1])](b[-1])
                    return aet[value_hash(b[-1])] <= (b[-1])
                else:
                    return AET.Serialized[value_hash(b[-1])] <= to_json(b[-1])
        elif isinstance(b[1], Delegate):
            return b[1]

        raise NotImplementedError(f"Unhandled type in dispatching {b}")

    for b in fg.blobs | filter[lambda b: b != None] | collect:
        el = dispatch_on_blob(b)
        if isinstance(el, LazyValue) or el != None: return_elements.add(el)

    from ..graph_delta import construct_commands
    return construct_commands(list(return_elements))




# ------------------------------FlatRef----------------------------------
def fr_source_imp(fr):
    assert isinstance(fr, FlatRef)
    blob = fr.fg.blobs[fr.idx]
    assert isinstance(blob[1], RelationType), f"| source is only allowed for FlatRef of type RelationType, but {type(blob[1])} was passed!"
    return FlatRef(fr.fg, blob[4])

def fr_target_imp(fr):
    assert isinstance(fr, FlatRef)
    blob = fr.fg.blobs[fr.idx]
    assert isinstance(blob[1], RelationType), f"| target is only allowed for FlatRef of type RelationType, but {type(blob[1])} was passed!"
    return FlatRef(fr.fg, blob[5])

def fr_value_imp(fr):
    assert isinstance(fr, FlatRef)
    blob = fr.fg.blobs[fr.idx]
    assert isinstance(blob[1], AttributeEntityType), "Can only ask for the value of an AET"
    return blob[-1]

def traverse_flatref_imp(fr, rt, direction, traverse_type):
    translation_dict = {
        "outout": "Out",
        "out"  : "out_rel",
        "inin": "In",
        "in": "in_rel",
    }
    assert isinstance(rt, RelationType), f"Passed Argument to traverse should be of type RelationType but got {rt}"
    blob = fr.fg.blobs[fr.idx]
    specific = blob[2] | filter[greater_than[0] if direction in {"out", "outout"} else less_than[0]] | collect
    specific_blobs = specific | map[lambda idx: fr.fg.blobs[abs(idx)]] | filter[lambda b: b[1] == rt] | collect
    if traverse_type == "single" and len(specific_blobs) != 1: return Error.ValueError(f"There isn't exactly one {translation_dict[direction]} RT.{rt} Relation. Did you mean {translation_dict[direction]}s[RT.{rt}]?")
    
    if direction == "inin": idx = 4
    elif direction == "outout": idx = 5
    else: idx = 0 # itself the relation

    if traverse_type == "single": return FlatRef(fr.fg, specific_blobs[0][idx])
    return FlatRefs(fr.fg, specific_blobs | map[lambda b: b[idx]] | collect)

def fr_outs_imp(fr):
    assert isinstance(fr, FlatRef)
    blob = fr.fg.blobs[fr.idx]
    outs = blob[2] | filter[greater_than[-1]] | collect
    return FlatRefs(fr.fg, outs)

def fr_ins_imp(fr):
    assert isinstance(fr, FlatRef)
    blob = fr.fg.blobs[fr.idx]
    ins = blob[2] | filter[less_than[0]] | collect
    return FlatRefs(fr.fg, ins)

def fr_ins_and_outs_imp(fr):
    assert isinstance(fr, FlatRef)
    blob = fr.fg.blobs[fr.idx]
    return FlatRefs(fr.fg, blob[2])

def fg_all_imp(fg, selector=None):
    assert is_a(fg, FlatGraph)
    if selector:
        selected_blobs = fg.blobs | filter[lambda b: is_a(b[1], selector)] | collect
    else:
        selected_blobs = fg.blobs 
    return FlatRefs(fg, [b[0] for b in selected_blobs])


# ------------------------------Merging FlatGraphs----------------------------------

def fg_merge_imp(fg1, fg2 = None):
    if isinstance(fg1, list): return fg1[1:] | reduce[fg_merge_imp][fg1[0]] | collect

    def idx_generator(n):
        def next_idx():
            nonlocal n
            n = n + 1
            return n
        return next_idx

    blobs, k_dict = [*fg1.blobs], {**fg1.key_dict}
    next_idx = idx_generator(length(blobs) - 1)

    idx_key_2 = {i:k for k,i in fg2.key_dict.items()}
    old_to_new = {}

    def retrieve_or_insert_blob(new_b):
        old_idx = new_b[0]
        key = idx_key_2.get(new_b[0], None)

        if old_idx in old_to_new:
            idx = old_to_new[old_idx]
            new_b = blobs[idx]
        elif (new_b[1] == 'BT.ValueNode' or is_a(new_b[3], uid)) and key in k_dict:
            new_b = blobs[k_dict[key]]
            idx = new_b[0]
        else:
            idx = next_idx()
            if key: k_dict[key] = idx
            new_b = (idx, new_b[1], [], *new_b[3:])
            blobs.append(new_b)
        old_to_new[old_idx] = idx
        return new_b

    for b in fg2.blobs | filter[lambda b: isinstance(b[1], RelationType)] | sort[lambda b: -len(b[2])]:
        rt_key = idx_key_2.get(b[0], None)

        src_b, trgt_b = fg2.blobs[b[4]], fg2.blobs[b[5]]
        src_b  = retrieve_or_insert_blob(src_b)
        trgt_b = retrieve_or_insert_blob(trgt_b)
        
        idx = next_idx()
        rt_b = (idx, b[1], [], None, src_b[0], trgt_b[0])
        src_b[2].append(idx)
        trgt_b[2].append(-idx)
        if rt_key: k_dict[rt_key] = idx
        blobs.append(rt_b)
        old_to_new[b[0]] = idx

            
    for b in fg2.blobs | filter[lambda b: not isinstance(b[1], RelationType)]:
        if b[0] not in old_to_new: retrieve_or_insert_blob(b)

    new_fg = FlatGraph()
    new_fg.blobs = blobs
    new_fg.key_dict = k_dict
    return new_fg



# This file was generated by 'versioneer.py' (0.22) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2023-03-08T17:19:55+1100",
 "dirty": false,
 "error": null,
 "full-revisionid": "99731ea365df1a59de674ca3b6326727289ee881",
 "version": "0.16.3.post11"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest  # pytest takes ages to run anything as soon as anything from zef is imported
from zef import *
from zef.ops import *

class MyTestCase(unittest.TestCase):
    def test_http_server(self):
        from zef.core.fx.http import send_response, permit_cors, middleware, middleware_worker, fallback_not_found, route
        port = 4990
        payload = b"successful test"
        eff = {
            'type': FX.HTTP.StartServer,
            # We hope that this is not taken by anything else for this test...
            'port': port,
            'pipe_into': (map[middleware_worker[permit_cors,
                                                route["/fixed-url"][lambda query: {"type": FX.HTTP.SendResponse,
                                                                                   "server_uuid": query["server_uuid"],
                                                                                   "request_id": query["request_id"],
                                                                                   "response": "Fixed"}],
                                                route["/test-url"][lambda query: {**query, "response_body": payload}],
                                                fallback_not_found,
                                                send_response]]
                          | subscribe[run]),
            'bind_address': "localhost",
        }

        eff_resp = eff | run

        import requests
        r = requests.get(f"http://localhost:{port}/test-url")
        self.assertEqual(r.status_code, 200)
        self.assertEqual(r.content, payload)

        r = requests.get(f"http://localhost:{port}/fixed-url")
        self.assertEqual(r.status_code, 200)
        self.assertEqual(r.content, b"Fixed")

        r = requests.get(f"http://localhost:{port}")
        self.assertEqual(r.status_code, 404)

        {
            "type": FX.HTTP.StopServer,
            "server_uuid": eff_resp["server_uuid"],
        } | run

if __name__ == '__main__':
    unittest.main()



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from zef import *


class TestStringMethods(unittest.TestCase):

    def test_blobtype_name(self):
        g = Graph()
        et = instantiate(ET.Entity, g)
        rt = instantiate(et, RT.Relation, et, g)
        aet = instantiate(AET.String, g)

        self.assertEqual(str(BT(et)), 'ENTITY_NODE')
        self.assertEqual(str(BT(aet)), 'ATTRIBUTE_ENTITY_NODE')
        self.assertEqual(str(BT(rt)), 'RELATION_EDGE')

        self.assertEqual(str(ET(et)), 'Entity')
        self.assertEqual(repr(ET(et)), 'ET.Entity')

        self.assertEqual(str(AET(aet)), 'String')
        self.assertEqual(repr(AET(aet)), 'AET.String')

        self.assertEqual(str(RT(rt)), 'Relation')
        self.assertEqual(repr(RT(rt)), 'RT.Relation')

        aet2 = instantiate(AET.QuantityFloat.kilograms, g)
        self.assertEqual(repr(AET(aet2)), 'AET.QuantityFloat.kilograms')
        self.assertEqual(str(AET(aet2)), 'QuantityFloat.kilograms')

        en = EN.Unit.kilograms
        self.assertEqual(repr(en), 'EN.Unit.kilograms')
        self.assertEqual(en.enum_type, "Unit")
        self.assertEqual(en.enum_value, "kilograms")


if __name__ == '__main__':
    unittest.main()



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import structlog

# Trying to keep a config that doesn't interfer with any user's global settings
# for structlog if they happen to use it.
import os
logger_type = os.environ.get("ZEFDB_LOGGER", "RICH").upper()
if logger_type not in ["RICH","PLAIN"]:
    print(f"Did not understand ZEFDB_LOGGER type of {logger_type}, going to use PLAIN by default.")
    logger_type = "PLAIN"
if logger_type == "RICH":
    log = structlog.wrap_logger(None,
        processors=[
            # structlog.stdlib.filter_by_level,
            # structlog.stdlib.add_logger_name,
            structlog.stdlib.add_log_level,
            # structlog.stdlib.PositionalArgumentsFormatter(),
            structlog.processors.TimeStamper(fmt="iso"),
            # The following only does anything if exc_info is passed as a keyword
            # structlog.processors.format_exc_info,
            structlog.processors.UnicodeDecoder(),
            structlog.dev.ConsoleRenderer(colors=True),
        ]
    )
elif logger_type == "PLAIN":
    log = structlog.wrap_logger(None,
        processors=[
            structlog.stdlib.add_log_level,
            structlog.processors.TimeStamper(fmt="iso"),
            # The following only does anything if exc_info is passed as a keyword
            # structlog.processors.format_exc_info,
            structlog.processors.UnicodeDecoder(),
            # structlog.processors.KeyValueRenderer(),
            structlog.dev.ConsoleRenderer(colors=False, exception_formatter=structlog.dev.plain_traceback),
        ]
    )
else:
    raise Exception("Should never get here")


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest  # pytest takes ages to run anything as soon as anything from zef is imported
from zef import *
from zef.ops import *
import zef


class MyTestCase(unittest.TestCase):

    def test_basics(self):
        g = Graph()

        z = ET.Machine | g | run

        self.assertEqual(g | blueprint | length | collect, 2)

        a,b,c = (z, RT.Something, 5) | g | run

        self.assertEqual(g | blueprint | length | collect, 5)

        self.assertEqual(g | now | blueprint | length | collect, 5)
        self.assertEqual(z | frame | blueprint | length | collect, 2)

        dz = delegate_of(z)
        db = delegate_of(b)

        with self.assertRaises(Exception):
            zef.pyzef.zefops.retire(to_ezefref(dz))
        with self.assertRaises(Exception):
            zef.pyzef.zefops.retire(to_ezefref(db))

        # Retire a node
        z | terminate | g | run

        zef.pyzef.zefops.retire(to_ezefref(db))

        self.assertEqual(g | now | blueprint | length | collect, 4)

        zef.pyzef.zefops.retire(to_ezefref(dz))
        self.assertEqual(g | now | blueprint | length | collect, 3)

    def test_meta(self):
        g = Graph()

        z = ET.Machine | g | run

        dz = to_ezefref(delegate_of(z))
        a,b,c = (dz, RT.Meta, 1) | g | run

        db = to_ezefref(delegate_of(b))

        with self.assertRaises(Exception):
            zef.pyzef.zefops.retire(dz)
        with self.assertRaises(Exception):
            zef.pyzef.zefops.retire(db)

        z | terminate | g | run

        with self.assertRaises(Exception):
            zef.pyzef.zefops.retire(dz)
        with self.assertRaises(Exception):
            zef.pyzef.zefops.retire(db)

        b | terminate | g | run

        with self.assertRaises(Exception):
            zef.pyzef.zefops.retire(dz)
        zef.pyzef.zefops.retire(db)

        with self.assertRaises(Exception):
            zef.pyzef.zefops.retire(dz)
        zef.pyzef.zefops.retire(delegate_of(dz))

        zef.pyzef.zefops.retire(dz)

    def test_recreate_delegate(self):
        g = Graph()

        z = ET.Machine | g | run

        self.assertEqual(g | now | blueprint | length | collect, 2)

        z | terminate | g | run
        zef.pyzef.zefops.retire(to_ezefref(delegate_of(z)))

        self.assertEqual(g | now | blueprint | length | collect, 1)

        z = ET.Machine | g | run

        self.assertEqual(g | now | blueprint | length | collect, 2)



if __name__ == '__main__':
    unittest.main()



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..core import *
from ..ops import *

sample_json = {
    "auth": {
        "JWKURL": "https://www.googleapis.com/service_accounts/v1/jwk/securetoken@system.gserviceaccount.com",
        "Header": "X-Auth-Token",
        "Audience": "ikura-9437c",
    },
    
    "types": {
        "User": {
            "firebaseID": {"type": "String",
                           "required": True,
                           "unique": True},

            "email": {"type": "String",
                      "required": True,
                      "unique": True},
            
            "givenName": {"type": "String"},
            
            "familyName": {"type": "String"},
            
            "phone": {"type": "String"},
            
            "birthday": {"type": "DateTime"},

            "active": {"type": "Boolean",
                       "search": True},
            
            "transactions": {"type": "Transaction",
                             "list": True,
                             "relation": (ET.Transaction, RT.User, ET.User),
                             "incoming": True},
            
            "categories": {"type": "Category",
                             "list": True,
                             "relation": (ET.Category, RT.User, ET.User),
                             "incoming": True},
        },
        
        "Transaction": {
            # Allow query only if user id matches graph data
            "_AllowQuery": "z >> RT.User >> RT.FirebaseID | value | equals[info.context['auth']['user_id']]",
            # Allow add only if user id matches incoming payload
            # This is currently available - could make this better (i.e. the same style as AllowQuery)
            "_AllowAdd": "params['input']['user']['id'] == info.context['auth']['user_id']",
            # Allow update only if user id matches both the graph data and the incoming payload (or is not overwritten in the incoming payload)
            "_AllowUpdate": """(Z >> RT.User >> RT.FirebaseID | value | equals[info.context['auth']['user_id']] | collect)
            and
            (get_path(params, ['input', 'user', 'id'], None) in [None, info.context['auth']['user_id']])
            """,
            # Note: AllowDelete falls back to AllowUpdate, with an empty payload
            
            "user": {"type": "User"},
            
            "type": {"type": "TransactionType",
                     "search": True},
            
            "amount": {"type": "Float",
                       "relation": (ET.Transaction, RT.Amount, AET.Float),
                       "search": True},
            
            "when": {"type": "DateTime",
                     "search": True},
            
            "category": {"type": "Category", "relation": (ET.Transaction, RT.Category, ET.Category)},
            
            "comment": {"type": "String",
                        "relation": (ET.Transaction, RT.Comment, AET.String)},
        },

        "Category": {
            # # Allow query only if user id matches graph data
            # "_AllowQuery": "z >> RT.User >> RT.ID | value | equals[info.context['auth']['user_id']",
            # # Allow add only if user id matches incoming payload
            # # This is currently available - could make this better (i.e. the same style as AllowQuery)
            # "_AllowAdd": "params['input']['user']['id'] == info.context['auth']['user_id']",
            # # Allow update only if user id matches both the graph data and the incoming payload (or is not overwritten in the incoming payload)
            # "_AllowUpdate": """(Z >> RT.User >> RT.ID | value | equals[info.context['auth']['user_id'] | collect
            # and
            # (get_path(params, ['input', 'user', 'id'], None) in [None, info.context['auth']['user_id']])
            # """,
            # # Note: AllowDelete falls back to AllowUpdate, with an empty payload
            
            "type": {"type": "CategoryType",
                     "search": True},

            "user": {"type": "User"},
            
            "title": {"type": "String",
                      "required": True},

            "titleLangEn": {"type": "String"},
            "titleLangJa": {"type": "String"},
            "description": {"type": "String"},
            "descriptionLangEn": {"type": "String"},
            "descriptionLangJa": {"type": "String"},
            "color": {"type": "String"},
            "icon": {"type": "String"},
            
            "createdAt": {"type": "DateTime",
                          "required": True},
            
            "transactions": {"type": "Transaction",
                             "list": True,
                             "relation": (ET.Transaction, RT.Category, ET.Category),
                             "incoming": True},
        },
    },

    "enums": {
        "TransactionType": [
            "EXPENSE",
            "INCOME"
        ],

        "CategoryType": [
            "DEFAULT",
            "PRIVATE"
        ],
    },
}


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from ..core.VT import ValueType_
from functools import partial


def zef_ui_ctor(type_name, self, *args, **kwargs):
    data = None
    if len(args) == 1: 
        data = args[0]
    elif len(args) > 1:
        raise ValueError(f'{type_name} constructor takes at most one positional argument, got {args}')

    if data:
        if "data" in kwargs: raise ValueError(f'{type_name} was passed both a data positional arg and a data key in kwargs which isn\'t allowed')
        else: kwargs = {**kwargs, "data": data}
    
    ctor_dict = {"type_name":type_name, "constructor_func": partial(zef_ui_ctor, type_name), "pass_self": True}
    if len(self.d['absorbed']) == 1:
        return ValueType_( absorbed = ({**self.d['absorbed'][0], **kwargs},), **ctor_dict)

    return ValueType_(absorbed = (kwargs,), **ctor_dict)



Text  = ValueType_(type_name='Text', constructor_func = partial(zef_ui_ctor, 'Text'), pass_self = True)
Code  = ValueType_(type_name='Code', constructor_func = partial(zef_ui_ctor, 'Code'), pass_self = True)
Style = ValueType_(type_name='Style', constructor_func = partial(zef_ui_ctor, 'Style'), pass_self = True)
Table = ValueType_(type_name='Table', constructor_func = partial(zef_ui_ctor, 'Table'), pass_self = True)
Column = ValueType_(type_name='Column', constructor_func = partial(zef_ui_ctor, 'Column'), pass_self = True)
Frame = ValueType_(type_name='Frame', constructor_func = partial(zef_ui_ctor, 'Frame'), pass_self = True)
HStack = ValueType_(type_name='HStack', constructor_func = partial(zef_ui_ctor, 'HStack'), pass_self = True)
VStack = ValueType_(type_name='VStack', constructor_func = partial(zef_ui_ctor, 'VStack'), pass_self = True)
BulletList   = ValueType_(type_name='BulletList', constructor_func = partial(zef_ui_ctor, 'BulletList'), pass_self = True)
NumberedList = ValueType_(type_name='NumberedList', constructor_func = partial(zef_ui_ctor, 'NumberedList'), pass_self = True)
Paragraph = ValueType_(type_name='Paragraph', constructor_func = partial(zef_ui_ctor, 'Paragraph'), pass_self = True)


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest  # pytest takes ages to run anything as soon as anything from zef is imported
from zef import *
from zef.ops import *
import zef

class MyTestCase(unittest.TestCase):
    def test_local_graph(self):
        import tempfile
        with tempfile.TemporaryDirectory() as tmpdir:
            # Creating new graph
            filepath = "file://" + os.path.join(tmpdir, "local.graph")
            g = Graph(filepath)
            created_uid = uid(g)
            z = ET.Machine | g | run

            zef.pyzef.main.save_local(g)

            # Unload graph so we can test the reload
            g = None
            import gc
            gc.collect()

            g = Graph(filepath)
            self.assertEqual(g | now | all[ET] | length | collect, 1)
            self.assertEqual(uid(g), created_uid)


if __name__ == '__main__':
    unittest.main()



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest  # pytest takes ages to run anything as soon as anything from zef is imported
from zef.ops import *
from zef import *

class MyTestCase(unittest.TestCase):
    def test_info(self):
        g = Graph()

        instantiate(AET.String, g)
        instantiate(AET.String, g)
        f = instantiate(AET.String, g)

        w = instantiate(ET.Worker, g)

        assign(f, 'Fritz')
        assign(f, 'Ninja')

        instantiate(w, RT.YearOfBirth, f, g)
        assign(f, 'Ninja2')
        assign(f, 'Ninja')
        assign(f, 'Yolandi')


        man = instantiate(ET.Manager, g)
        m = instantiate(ET.Machine, g)
        m2 = instantiate(ET.Machine, g)

        m3 = instantiate(ET.Machine, g)
        n = instantiate(AET.String, g)


        instantiate(w, RT.CanOperate, m, g)
        instantiate(w, RT.Name, n, g)
        instantiate(w, RT.CanOperate, m2, g)
        rr = instantiate(w, RT.CanOperate, m3, g)
        instantiate(w, RT.CanOperate, m3, g)

        can_op = instantiate(man, RT.CanOperate, w, g)

        terminate(f) | g | run
        terminate(w) | g | run

        rr | to_frame[can_op | frame | collect] | yo[False] | collect
        rr | to_ezefref | yo[False] | collect

if __name__ == '__main__':
    unittest.main()



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#=============================================================================== WS handlers =========================================================================================================
from .fx_types import Effect

def websocket_connect_to_server_handler(eff: Effect):
    raise NotImplementedError("websocket_connect_to_server_handler not implemented")

def websocket_start_server_handler(eff: Effect):
    raise NotImplementedError("websocket_start_server_handler not implemented")

def websocket_stop_server_handler(eff: Effect):
    raise NotImplementedError("websocket_stop_server_handler not implemented")

def websocket_send_message_handler(eff: Effect):
    raise NotImplementedError("websocket_send_message_handler not implemented")

def websocket_close_connections_handler():
    raise NotImplementedError("websocket_close_connections_handler not implemented")

